/*
  Copyright (c) 2008 Andres Colubri

  This source is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This code is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  General Public License for more details.

  A copy of the GNU General Public License is available on the World
  Wide Web at <http://www.gnu.org/copyleft/gpl.html>. You can also
  obtain it by writing to the Free Software Foundation,
  Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

// This class implements a bakground differencing algorithm to detect changes in the
// image captured by the camera.
public class Detector
{
    Detector(PApplet parent, int w, int h, float th, int f0, int rx, int ry, float ff)
    {
	img = new PImage(w, h); 
        img0 = new PImage(w, h);
        imgDiff = new PImage(w, h);

        threshold = th; 
        firstFrame = f0;
        resx = rx;
        resy = ry;
        freeFraction = ff;
        
        diffGrid = new boolean[resx][resy];
        initDiffGrid();
        
        frame = 0;
    }
    
    // Uses the image provided in cam to run the background differencing.
    void run(PImage cam)
    {
        img.copy(cam, 0, 0, cam.width, cam.height, 
                      0, 0, img.width, img.height);
        fastBlur(); // Applies a fast blur to the image so the effect of noise is minimized.

        if (diffFirstFrame <= frame)
            if (frame == diffFirstFrame) initImg0(); // Initializing background.
            else
            { 
                computeImgDiff();
                cleanImgDiff();
            }

        frame++;
    }
    
    // Updates the wall, by "hitting" those regions in the wall
    // where a motion has been detected.
    void update(Wall wall)
    {
        for (int i = 0; i < resx; i++)
            for (int j = 0; j < resy; j++)
                if (diffGrid[i][j]) wall.hit(i, j);
    }
    void renderImg()
    {
        img.updatePixels();
        image(img, 0, 0, width, height);  
    }
    
    void renderImg0()
    {
        image(img0, 0, 0, width, height);  
    }    
    
    void renderImgDiff()
    {
        imgDiff.updatePixels();
        image(imgDiff, 0, 0, width, height);  
    }
    
    void restart()
    {
        initDiffGrid();
        frame = 0;  
    }
    
    // ==================================================
    // Super Fast Blur v1.1
    // by Mario Klingemann 
    // <http://incubator.quasimondo.com>
    // ==================================================
    private void fastBlur()
    {
        int radius = 2;
        int w = img.width;
        int h = img.height;
        int wm = w - 1;
        int hm = h - 1;
        int wh = w * h;
        int div = radius + radius + 1;
        int r[] = new int[wh];
        int g[] = new int[wh];
        int b[] = new int[wh];
        int rsum, gsum, bsum, x, y, i, p, p1, p2, yp, yi, yw;
        int vmin[] = new int[max(w, h)];
        int vmax[] = new int[max(w, h)];
        int[] pix = img.pixels;
        int dv[] = new int[256*div];
    
        for (i = 0; i < 256 * div; i++)
        {
            dv[i] = (i / div);
        }

        yw = yi = 0;

        for (y = 0; y < h; y++)
        {
            rsum = gsum = bsum = 0;
            
            for (i = -radius; i<=radius; i++)
            {
              p = pix[yi + min(wm, max(i,0))];
              rsum += (p & 0xff0000) >> 16;
              gsum += (p & 0x00ff00) >> 8;
              bsum += p & 0x0000ff;
            }
            
            for (x = 0; x < w; x++)
            {
                r[yi] = dv[rsum];
                g[yi] = dv[gsum];
                b[yi] = dv[bsum];

                if (y == 0)
                {
                    vmin[x] = min(x + radius + 1, wm);
                    vmax[x] = max(x - radius, 0);
                }
                p1 = pix[yw + vmin[x]];
                p2 = pix[yw + vmax[x]];

                rsum += ((p1 & 0xff0000) - (p2 & 0xff0000)) >> 16;
                gsum += ((p1 & 0x00ff00) - (p2 & 0x00ff00)) >> 8;
                bsum += (p1 & 0x0000ff) - (p2 & 0x0000ff);
                yi++;
            }
            yw += w;
        }

        for (x = 0; x < w; x++)
        {
            rsum = gsum = bsum = 0;
            yp =- radius * w;
            for (i = -radius; i <= radius; i++)
            {
                yi = max(0, yp) + x;
                rsum += r[yi];
                gsum += g[yi];
                bsum += b[yi];
                yp += w;
            }
            
            yi = x;
            for (y = 0; y < h; y++)
            {
                pix[yi] = 0xff000000 | (dv[rsum] << 16) | (dv[gsum] << 8) | dv[bsum];
                if (x == 0)
                {
                    vmin[y] = min(y + radius + 1, hm) * w;
                    vmax[y] = max(y - radius, 0) * w;
                }
            
                p1 = x + vmin[y];
                p2 = x + vmax[y];

                rsum += r[p1] - r[p2];
                gsum += g[p1] - g[p2];
                bsum += b[p1] - b[p2];

                yi += w;
            }
        }
    }

    private void initImg0()
    {
        img.updatePixels();
        img0.copy(img, 0, 0, img.width, img.height, 
                       0, 0, img.width, img.height);
        img0.updatePixels();      
    }
    
    // The background differencing algorithm.
    private void computeImgDiff()
    {
        int p, r, g, b;
        int p0, r0, g0, b0;            
        float d;
            
        int[] pix = img.pixels;
        int[] pix0 = img0.pixels;
        int[] diff = imgDiff.pixels;            
        for (int i = 0; i < img.width * img.height; i++)
        {
            p = pix[i];
            r = (p & 0xff0000) >> 16;
            g = (p & 0x00ff00) >> 8;
            b = p & 0x0000ff;
              
            p0 = pix0[i];
            r0 = (p0 & 0xff0000) >> 16;
            g0 = (p0 & 0x00ff00) >> 8;
            b0 = p0 & 0x0000ff;
                
            // Difference between reference pixel and new pixel.    
            d = abs(r0 - r) + abs(g0 - g) + abs(b0 -b);
                
            if (d < threshold)
            {
                // The difference is not large enough.
                diff[i] = 0xff000000 | (0 << 16) | (0 << 8) | 0;
            }
            else 
            {
                // The difference is large enough to be considered as a possible motion.
                diff[i] = 0xff000000 | (255 << 16) | (255 << 8) | 255;
            }                
        }
    }     

    // It takes the (resx, resy) grid that represents the wall and makes sure that
    // inside each cell of the grid there are enough different pixels to consider
    // that there actually was motion inside that cell.
    private void cleanImgDiff()
    {
        float w = float(imgDiff.width) / resx;
        float h = float(imgDiff.height) / resy;
        
        float f;
        int x, y, x0, x1, y0, y1, n;
        int[] diff = imgDiff.pixels;
        for (int i = 0; i < resx; i++)
        {
            x0 = min(imgDiff.width - 1, int(i * w));
            x1 = min(imgDiff.width - 1, int(x0 + w));
            for (int j = 0; j < resy; j++)
            {
                y0 = min(imgDiff.height - 1, int(j * h));
                y1 = min(imgDiff.height - 1, int(y0 + h));
                
                // Counting how many different pixels in the (i, j) block.
                n = 0;
                for (x = x0; x <= x1; x++)
                    for (y = y0; y <= y1; y++)
                    {
                        if (diff[y * imgDiff.width + x] != 0xff000000) n++;
                    }
                
                f = float(n) / ((x1 - x0 + 1) * (y1 - y0 + 1));
                if (f < diffFreeFraction)
                {
                    if (0 < f)
                    {
                        // Too few different pixels (but at least one). Cleaning region.
                        for (x = x0; x <= x1; x++)
                            for (y = y0; y <= y1; y++)
                            {
                                diff[y * imgDiff.width + x] = 0xff000000;
                            }
                    }                
                    diffGrid[i][j] = false; // No motion at (i,j).
                }
                else diffGrid[i][j] = true; // Motion has been detected at (i,j).
            }      
        }
    }
    
    private void initDiffGrid()
    {
        for (int i = 0; i < resx; i++)
            for (int j = 0; j < resy; j++)
                diffGrid[i][j] = false;
    }

    private PImage img;

    private PImage img0, imgDiff;
    int frame;
    
    boolean[][] diffGrid;
    
    float threshold;
    int firstFrame;
    int resx, resy;
    float freeFraction;
}
