/**
 * Interactive Wall
 * by Andres Colubri. March 2008.
 * 
 * Maps motion detected by the camera onto a light wall controlled with an DMX USB Pro
 * adaptor. This has been used in the Superficial Superglow 2.0 project proposal for the
 * Architecture Tech Seminar thought by David Erdman during the Fall Quarter of 2008 at
 * the Architecture and Urban Design deptartment at UCLA.
 *
 * The program maps the DMX channels to a 2D array that represents the LED lights attached to 
 * wall. This allow to control any LED topology.
 * There are three lighting behaviours implemented in this version, which are: hyper, cool
 * and sleepy. They differ in the dimming curves used for controling the intensity of the LEDs.
 *
 * It reads the keyboard for 4 inputs: 'enter' restarts the detection algorithm (it takes the 
 * camera image as the moment of the keypress to be the new bakground for the background
 * differencing algorithm), meanwhile '1', '2' and '3' set the light behaviour to hyper, cool
 * and sleepy, respectively.
 *
 * Version 6.0
 * Added comments.
 */

/*
  Copyright (c) 2008 Andres Colubri

  This source is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This code is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  General Public License for more details.

  A copy of the GNU General Public License is available on the World
  Wide Web at <http://www.gnu.org/copyleft/gpl.html>. You can also
  obtain it by writing to the Free Software Foundation,
  Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

// GLOBAL PARAMETRERS *****************************************************

// In test mode, no DMX messages are sent out.
boolean TEST_MODE = true;

// The size of the screen.
int screenWidth = 800;
int screenHeight = 600;

// The framerate.
int renderFPS = 30;

boolean renderCam = false;        // If true, it renders the image from the camera.
boolean renderWall = true;        // If true, it renders the light intensities as they are send to the DMX.
boolean renderMask = true;        // If true, it renders the mask applied onto the lights.
boolean renderImgDiff = false;    // If true, it renders the bitmap containing the image difference used by the tracking algorithm.

// Resolution and framerate for the camera.
int camWidth = 320;
int camHeight = 240;
int camFPS = 15;

// Size of the wall grid. This means that a maximum of wallResX x wallResY will be handled.
int wallResX = 4;
int wallResY = 3;

// Factor to smooth out the changes in light intensity. wallSmoothFactor = 1 means no smoothing,
// wallSmoothFactor -> 1 means increased smoothing.
float wallSmoothFactor = 0.8;
// Timestep for the lights, in milliseconds.
int wallTimeStep = 100;
// Initial lighting behavior: start time, constant time, falloff time.
float wallLightStartTime = 0.05;
float wallLightConstTime = 0.1;
float wallLightFinishTime = 0.05;
// Minimum and maximum grey values used to display the lights in the screen.
int wallColorMinBright = 160;
int wallColorMaxBright = 255;

// Size of the image used to do the motion tracking calculations.
int diffWidth = 80;
int diffHeight = 60;
// Threshold for the differencing algorithm: the larger 
// this value is, the more insensitive the tracking becomes
// (that is, it would only detect big changes in the image).
float diffThreshold = 150.0;
// Frame when the tracking detection starts.
int diffFirstFrame = 30;
// If a region in the image has less than diffFreeFraction of different pixels
// (different with respect to the background image), then it is not regarded 
// as contain a motion. wallResX and wallResY are used to determine the number
// of regions in the image.
float diffFreeFraction = 0.1;

// DMX configuration: COM port to which the DMX USB adaptor is connected to, 
// bitrate and maximum number of channels.
String dmxPort = "COM3";
int dmxSpeed = 115200;
int maxChannelNum = 32;

String maskFile = "maskFull.png";

import JMyron.*;

boolean newFrame = true;

JMyron cam;
Detector det; 
Wall wall; 

PImage img;

PImage imgMask;

float camFrameTime;
int lastFrameTime;

void setup()
{
    size(screenWidth, screenHeight);
    smooth();
    colorMode(RGB, 255);
    frameRate(renderFPS);
    noCursor();

    det = new Detector(this, diffWidth, diffHeight, diffThreshold, diffFirstFrame, wallResX, wallResY, diffFreeFraction);    
    wall = new Wall(this, wallResX, wallResY, wallSmoothFactor, wallLightStartTime, wallLightConstTime, wallLightFinishTime, wallTimeStep, 
                          wallColorMinBright, wallColorMaxBright, dmxPort, dmxSpeed, maxChannelNum);
    cam = new JMyron();
    cam.start(camWidth, camHeight);
    cam.findGlobs(0); // Disable the intelligence to speed up frame rate.
    
    img = new PImage(camWidth, camHeight);
    img.loadPixels();
    
    camFrameTime = 1000.0 / camFPS;
    lastFrameTime = 0;
  
    imgMask = loadImage(maskFile);

    initDMXChannels();
}

void draw()
{ 
    int t = millis();
    if (t - lastFrameTime >= camFrameTime)
    {
        lastFrameTime = t;
        
        cam.update();               // Update the camera view.
        cam.imageCopy(img.pixels);  // Grab the camera pixels.
        img.updatePixels();

        det.run(img);               // Run motion detection.
        det.update(wall);           // Updates wall with new motions.
        wall.incTime();             // Increases wall clock.
        wall.update();              // Updates lights status.
    
        newFrame = true;
    }

    if (newFrame)
    {
        background(0);	

        wall.send();                // Sending light states to the DMX.
        if (renderCam) image(img, 0, 0, width, height);
        if (renderImgDiff) det.renderImgDiff();
        if (renderWall) wall.render();
        if (renderMask) image(imgMask, 0, 0, width, height);
        newFrame = false;
    }
}

// With this function the DMX channels are mapped to the 2D wall array.
void initDMXChannels()
{
    for (int i = 0; i < wallResX; i++)
        for (int j = 0; j < wallResY; j++)
        {
            wall.setChannel(i, j, -1);  
        }
        
    wall.setChannel(0, 0, 5); 
    
    wall.setChannel(1, 0, 4); 
    wall.setChannel(1, 1, 3);
    
    wall.setChannel(2, 1, 2);
    
    wall.setChannel(3, 0, 1);
    
    wall.setChannel(4, 1, 9); // Channel 0 is not working for some reason, seems to be a bug in the code...
    
    wall.setChannel(5, 0, 8);    

    wall.setChannel(6, 1, 7);
    
    wall.setChannel(7, 0, 6);    
}

void keyPressed()
{
    if (key == ENTER)
    {
        println("Restarting detector...");          
        det.restart();
    } 
    else if (key == '1')
    {
        println("Settting hyper mode...");
        wall.setBrightCurves(0.0, 0.05, 0.0);
    }
    else if (key == '2')
    {
        println("Settting cool mode...");
        wall.setBrightCurves(0.05, 0.1, 0.05);
    }
    else if (key == '3')
    {
        println("Settting sleepy mode...");
        wall.setBrightCurves(0.1, 0.2, 0.25);          
    }        
}
