/*
  Copyright (c) 2008 Andres Colubri

  This source is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This code is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  General Public License for more details.

  A copy of the GNU General Public License is available on the World
  Wide Web at <http://www.gnu.org/copyleft/gpl.html>. You can also
  obtain it by writing to the Free Software Foundation,
  Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

// This class encapsulates a 2D array that represents the light wall. Each cell in the array
// corresponds to a DMX channel that controls one light or set of lights through the DMX adapter.
public class Wall
{
    public Wall(PApplet parent, int rx, int ry, float f, float t0, float t1, float t2, float ts, int minb, int maxb, String dmxPort, int dmxSpeed, int maxChannelNum)
    {
        if (!TEST_MODE) dmx = new DMX(parent, dmxPort, dmxSpeed, maxChannelNum);
      
        resx = rx;
        resy = ry;
        smoothFactor = f;
        
        deltax = float(width) / rx;
        deltay = float(height) / ry;        
        
        brightGrid = new float[rx][ry];
        channelGrid = new int[rx][ry];
        hitTimeGrid = new int[rx][ry];        
        bcurveGrid = new BrightCurve[rx][ry];
 
        time = 0;
        invTimeStep = 1 / ts;
        
        colorMinBright = minb; 
        colorMaxBright = maxb;

        setNoChannels();        
        setBrightToZero();
        setHitTimeToZero();        
        initBrightCurves(t0, t1, t2);
    }
    
    // Sets the value of channel associated to cell (i,j) to c.
    void setChannel(int i, int j, int c)
    {
       if ((0 <= i) && (i < resx) && (0 <= j) && (j < resy)) 
           channelGrid[i][j] = c;
    }
    
    // Hits all the cells in the rect (x, y, x + w, y + h).
    // To hit a cell means to indicate that a motion has occurred inside the
    // region that corresponds to that cell.
    void hit(float x, float y, float w, float h)
    {
        int i0 = min(resx - 1, int(x * resx));
        int j0 = min(resy - 1, int(y * resy));
        int i1 = min(resx - 1, i0 + int(w * resx));
        int j1 = min(resy - 1, j0 + int(h * resy));
        
        int t = time;
        for (int i = i0; i <= i1; i++)
            for (int j = j0; j <= j1; j++)
                hitTimeGrid[i][j] = t;
    }
    
    // Hits the cell (i,j).
    void hit(int i, int j)
    {
        if ((0 <= i) && (i < resx) && (0 <= j) && (j < resy))
            hitTimeGrid[i][j] = time;
    }    
    
    // Writes the intensity values to the DMX.
    void send()
    {
        int iflip;
        int channel, value;
        for (int i = 0; i < resx; i++)
            for (int j = 0; j < resy; j++)
            {
                // Fliping x value so the motion is mirrored.
                iflip = resx - i - 1;                  
                channel = channelGrid[iflip][j];
                if (-1 < channel)
                {
                     value = int(255 * brightGrid[i][j]);
                     if (!TEST_MODE) dmx.setDMXChannel(channel, value);
                }
            }
    }
    
    // Draws the grid using the intensity values as gray values.
    void render()
    {
        color fillColor;
        noStroke();
        int iflip;
        for (int i = 0; i < resx; i++)
            for (int j = 0; j < resy; j++)
            {
                iflip = resx - i - 1;
                float b = map(brightGrid[i][j], 0, 1.0, colorMinBright, colorMaxBright);                
                fill(b);
                rect(iflip * deltax, j * deltay, deltax, deltay);
            }
    }
    
    // Updates the light states by setting the new hit times.
    void update()
    {
        float b0, b1;
        int t = time;
        for (int i = 0; i < resx; i++)
            for (int j = 0; j < resy; j++)
            {
                bcurveGrid[i][j].updateHitTime(hitTimeGrid[i][j]);
                b0 = brightGrid[i][j];
                b1 = bcurveGrid[i][j].calculate(t);
                brightGrid[i][j] += smoothFactor * (b1 - b0); // Using easign formula for smooth changes in intensity.
            }     
    }
    
    void incTime()
    {
        time++;
    }

    void setBrightCurves(float t0, float t1, float t2)
    {
        for (int i = 0; i < resx; i++)
            for (int j = 0; j < resy; j++)  
                bcurveGrid[i][j].setTimeDeltas(t0, t1, t2);
    }
    
    private void setNoChannels()
    {
        for (int i = 0; i < resx; i++)
            for (int j = 0; j < resy; j++)
                channelGrid[i][j] = -1;
    }
    private void setBrightToZero()
    {
        for (int i = 0; i < resx; i++)
            for (int j = 0; j < resy; j++)
                brightGrid[i][j] = 0;      
    }
    private void setHitTimeToZero()
    {
        for (int i = 0; i < resx; i++)
            for (int j = 0; j < resy; j++)
                hitTimeGrid[i][j] = 0;            
    }
    private void initBrightCurves(float t0, float t1, float t2)
    {
        for (int i = 0; i < resx; i++)
            for (int j = 0; j < resy; j++)  
                bcurveGrid[i][j] = new BrightCurve(t0, t1, t2, invTimeStep);
    }

    private float colorMinBright; 
    private float colorMaxBright;
    private int time;
    private float invTimeStep;
    private DMX dmx;
    private float smoothFactor;
    private int resx, resy;
    private float deltax, deltay;
    private float[][] brightGrid;
    private int [][] channelGrid;    
    private BrightCurve[][] bcurveGrid;
    private int[][] hitTimeGrid;
}
